/* Sonogram
 * Written by Emil Mikulic, 2002.
 * http://dmr.ath.cx/sound/sonogram/
 *
 * Feel free to do whatever you like with this code.
 * Feel free to credit me.
 */

#include <math.h>
#include <stdio.h>

/* http://www.zip.com.au/~erikd/libsndfile/ */
#include <sndfile.h>

/* http://dmr.ath.cx/gfx/targa/ */
#include "targa.h"

#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
/* overkill */
#  define PI 3.141592653589793238462643383279502884197169399375105820974
# endif
#endif


/* --- global settings --- */
#define INFILE		"guitar.wav"
#define OUTFILE		"sonogram.tga"
#define FFT_SIZE	65536/2
#define FRAMES		640		/* width */
#define HEIGHT		480		/* (FFT_SIZE / 2) */
#define WINDOW_FUNC	w_rect

/* (image brightness) */
#define LEVEL_G		0.1
#define LEVEL_B		0.01



/* --- windowing --- */
__inline double w_hann(double pos)
{
	double x = cos(pos*PI/2.0);
	return x*x;
}

__inline double w_rect(double pos)
{
	return 1.0;
}



/* --- FFT --- */
double fft_re[FFT_SIZE], fft_im[FFT_SIZE];
int power;

/* reverse bits according to POWER */
__inline int br(const int i)
{
	int j, k = 0;

	for (j=0; j<power; j++)
	if ((i >> j) & 1)
		k = k | (1<<(power-j-1));

	return k;
}

__inline void fft(void)
{
	int blocks, points, i;

	for (blocks=FFT_SIZE/2,points=2; blocks>=1; blocks/=2,points*=2)
	for (i=0; i<blocks; i++)
	{
		int j, offset = points * i;

		/* "butterfly" odd elements */
		for (j=0; j<points/2; j++)
		{
			double t_re, t_im, top_r, top_i, btm_r, btm_i;
			double theta = -2.0 * PI *
				(double)j / (double)(points);

			t_re = cos(theta);
			t_im = sin(theta);

			top_r = fft_re[j+offset];
			top_i = fft_im[j+offset];

			btm_r = fft_re[j+points/2+offset]*t_re -
				fft_im[j+points/2+offset]*t_im;

			btm_i = fft_re[j+points/2+offset]*t_im +
				fft_im[j+points/2+offset]*t_re;			

			fft_re[j+offset] = top_r + btm_r;
			fft_im[j+offset] = top_i + btm_i;

			fft_re[j+points/2+offset] = top_r - btm_r;
			fft_im[j+points/2+offset] = top_i - btm_i;
		}
	}
}



/* --- error handling --- */
__inline void ERR(int condition, char *reason)
{
	if (!condition) return;

	/* on condition, throw reason */
	fprintf(stderr, "%s\n", reason);
	exit(1);
}



/* --- graphics --- */
#define CLIP(x) ((x)<0.0?0.0: ((x)>1.0?1.0: (x)))
__inline void put_pixel(byte *pixel, double v)
{
	double x;

	x = 1.0 + CLIP(v) * 9.0;
	x = log10(x);
	pixel[0] = 8 + (byte)(x * 247.0);

	x = 1.0 + CLIP(v/LEVEL_G) * 9.0;
	x = log10(x);
	pixel[1] = 8 + (byte)(x * 247.0);

	x = 1.0 + CLIP(v/LEVEL_B) * 9.0;
	x = log10(x);
	pixel[2] = (byte)(x * 255.0);
}
#undef CLIP



/* --- body program --- */
int main()
{
	int		i, j, skip;
	double		*tmp, loudest = 0.0, *table;
	SF_INFO		sfinfo;
	SNDFILE		*sndfile;
	tga_image	img;

	power = (int)(log((double)FFT_SIZE) / log(2.0));
	ERR((1<<power != FFT_SIZE), "FFT_SIZE not a power of two!");

	sndfile = sf_open_read(INFILE, &sfinfo);
	printf("Opened [%s] for reading\n\n"
		"samplerate:  %d\n"
		"samples:     %d\n"
		"channels:    %d\n"
		"pcmbitwidth: %d\n\n",
		INFILE,
		sfinfo.samplerate, sfinfo.samples,
		sfinfo.channels, sfinfo.pcmbitwidth);
	ERR((sfinfo.channels != 1), "Need mono input.");

	/* Init */
	tmp = (double*)malloc(sizeof(double) * sfinfo.samples);
	table = (double*)malloc(sizeof(double) * FRAMES * FFT_SIZE/2);
	ERR((!tmp || !table), "Out of memory!");

	skip = (sfinfo.samples + FFT_SIZE) / FRAMES;
	printf("%.2f%% overlap (skip = %d)\n",
		(double)(FFT_SIZE - skip) / (double)FFT_SIZE * 100.0,
		skip);

	/* Read entire sample into memory */
	ERR((sf_read_double(sndfile, tmp, sfinfo.samples, 1)
		!= sfinfo.samples), "Failed to read all samples.");
	sf_close(sndfile);

	/* Process frames */
	for (i=0; i<FRAMES; i++)
	{
		int start = -FFT_SIZE + i*skip;

		/* Sort into bit-reversed order */
		for (j=0; j<FFT_SIZE; j++)
		{
			int pos = br(j);

			fft_im[pos] = 0;
			if ((j+start < 0) || (j+start >= sfinfo.samples))
				fft_re[pos] = 0.0;
			else
			fft_re[pos] = tmp[j+start] *
				WINDOW_FUNC(
				(double)(j-FFT_SIZE/2) / (double)(FFT_SIZE/2)
				);
		}

		/* (hooray) */
		fft();

		/* Update 'loudest' amplitude for normalization */
		for (j=0; j<FFT_SIZE/2; j++)
		{
			double amp, re, im;

			re = fft_re[j];
			im = fft_im[j];
			amp = sqrt(re*re + im*im) / (double)FFT_SIZE;
			loudest = max(loudest, amp);

			table[j*FRAMES + i] = amp;
		}

		printf("Done %d/%d\r", i+1, FRAMES);
	}
	printf("\nLoudest: %f\nDrawing to file...", loudest);
	fflush(stdout);
	
	/* draw */
	img.width = FRAMES;
	img.height = HEIGHT;
	img.bits_per_pixel = 24;
	img.color_order = RGB;
	img.line_order = BOTTOM_TO_TOP;
	img.data = (byte*)malloc(FRAMES*FFT_SIZE/2*3);

	for (j=0; j<HEIGHT; j++)
	{
		int ofs = 3*j*FRAMES;

		for (i=0; i<FRAMES; i++)
		{
			double amp = table[j*FRAMES + i] / loudest;
			put_pixel(img.data + ofs, amp);
			ofs += 3;	
		}
	}
	tga_write(OUTFILE, &img);
	free(tmp);
	free(table);
	free(img.data);

	printf("done!\n");
	return 0;
}
